from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING

from pytest import fixture, mark

from utilities.platform import IS_NOT_LINUX
from utilities.sqlalchemy import ensure_tables_dropped_async

if TYPE_CHECKING:
    from collections.abc import Callable

    from sqlalchemy import Engine
    from sqlalchemy.ext.asyncio import AsyncEngine

    from utilities.asyncio import Coroutine1
    from utilities.sqlalchemy import TableOrMappedClass

FLAKY = mark.flaky(reruns=5, reruns_delay=1)
SKIPIF_CI = mark.skipif("CI" in environ, reason="Skipped for CI")
SKIPIF_CI_AND_NOT_LINUX = mark.skipif(
    ("CI" in environ) and IS_NOT_LINUX, reason="Skipped for CI/non-Linux"
)


# hypothesis

try:
    from utilities.hypothesis import setup_hypothesis_profiles
except ModuleNotFoundError:
    pass
else:
    setup_hypothesis_profiles()


# sqlalchemy


try:
    pass
except ModuleNotFoundError:
    pass
else:

    @fixture(scope="session")
    def create_postgres_engine() -> Callable[..., Engine]:
        """Create a Postgres engine."""

        def inner(*tables_or_mapped_classes: TableOrMappedClass) -> Engine:
            from utilities.sqlalchemy import create_engine, ensure_tables_dropped

            engine = create_engine(
                "postgresql", host="localhost", port=5432, database="testing"
            )
            ensure_tables_dropped(engine, *tables_or_mapped_classes)
            return engine

        return inner

    @fixture(scope="session")
    def create_postgres_engine_async() -> Callable[..., Coroutine1[AsyncEngine]]:
        """Create a Postgres engine."""

        async def inner(*tables_or_mapped_classes: TableOrMappedClass) -> AsyncEngine:
            from utilities.sqlalchemy import create_engine

            engine = create_engine(
                "postgresql+asyncpg",
                host="localhost",
                port=5432,
                database="testing",
                async_=True,
            )
            await ensure_tables_dropped_async(engine, *tables_or_mapped_classes)
            return engine

        return inner



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from utilities.typed_settings import click_field


@dataclass(frozen=True)
class Config:
    """Settings for the `pypi_server` script."""

    path_password: Path = click_field(
        default=Path("password"),
        help="generate using 'htpasswd -nbB username password'",
        param_decls=("-pw", "--path-password"),
    )
    path_packages: Path = click_field(
        default=Path("packages"), param_decls=("-pk", "--path-packages")
    )
    port: int = click_field(default=1461, param_decls=("-po", "--port"))
    dry_run: bool = click_field(default=False, param_decls=("-dr", "--dry-run"))
    exist_ok: bool = click_field(default=False, param_decls=("-e", "--exist-ok"))



from __future__ import annotations

from math import inf, nan
from typing import TYPE_CHECKING, ClassVar, Literal

import redis
import redis.asyncio
from hypothesis import HealthCheck, assume, given, settings
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    data,
    datetimes,
    floats,
    sampled_from,
    tuples,
)
from polars import Boolean, DataFrame, DataType, Float64, Int64, Utf8
from polars.testing import assert_frame_equal
from pytest import mark, param, raises
from redis.commands.timeseries import TimeSeries

from tests.conftest import FLAKY, SKIPIF_CI_AND_NOT_LINUX
from utilities.datetime import EPOCH_NAIVE, EPOCH_UTC, drop_microseconds
from utilities.hypothesis import (
    int32s,
    lists_fixed_length,
    redis_clients,
    redis_time_series,
    text_ascii,
    zoned_datetimes,
)
from utilities.polars import DatetimeUTC, check_polars_dataframe, zoned_datetime
from utilities.redis import (
    TimeSeriesAddDataFrameError,
    TimeSeriesAddError,
    TimeSeriesMAddError,
    TimeSeriesRangeError,
    TimeSeriesReadDataFrameError,
    ensure_time_series_created,
    time_series_add,
    time_series_add_dataframe,
    time_series_get,
    time_series_madd,
    time_series_range,
    time_series_read_dataframe,
    yield_client,
    yield_client_async,
    yield_time_series,
    yield_time_series_async,
)
from utilities.zoneinfo import HONG_KONG, UTC

if TYPE_CHECKING:
    import datetime as dt
    from uuid import UUID
    from zoneinfo import ZoneInfo

    from polars._typing import PolarsDataType, SchemaDict

    from utilities.types import Number


def _clean_datetime(
    datetime: dt.datetime, /, *, time_zone: ZoneInfo | None = None
) -> dt.datetime:
    _ = assume(datetime.fold == 0)
    if time_zone is not None:
        datetime = datetime.replace(tzinfo=time_zone)
    return max(datetime, EPOCH_UTC.astimezone(datetime.tzinfo))


@SKIPIF_CI_AND_NOT_LINUX
class TestEnsureTimeSeriesCreated:
    @FLAKY
    @given(client_pair=redis_clients(), key=text_ascii())
    def test_main(self, *, client_pair: tuple[redis.Redis, UUID], key: str) -> None:
        client, uuid = client_pair
        full_key = f"{uuid}_{key}"
        assert client.exists(full_key) == 0
        for _ in range(2):
            ensure_time_series_created(client.ts(), full_key)
        assert client.exists(full_key) == 1


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesAddAndGet:
    @FLAKY
    @given(
        ts_pair=redis_time_series(),
        key=text_ascii(),
        timestamp=datetimes(
            min_value=EPOCH_NAIVE, timezones=sampled_from([HONG_KONG, UTC])
        ).map(drop_microseconds),
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    def test_main(
        self,
        *,
        ts_pair: tuple[TimeSeries, UUID],
        key: str,
        timestamp: dt.datetime,
        value: float,
    ) -> None:
        ts, uuid = ts_pair
        full_key = f"{uuid}_{key}"
        timestamp = _clean_datetime(timestamp)
        res_add = time_series_add(
            ts, full_key, timestamp, value, duplicate_policy="last"
        )
        assert isinstance(res_add, int)
        res_timestamp, res_value = time_series_get(ts, full_key)
        assert res_timestamp == timestamp.astimezone(UTC)
        assert res_value == value

    @given(
        ts_pair=redis_time_series(),
        key=text_ascii(),
        timestamp=zoned_datetimes(min_value=EPOCH_NAIVE).map(drop_microseconds),
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    def test_error_at_upsert(
        self,
        *,
        ts_pair: tuple[TimeSeries, UUID],
        key: str,
        timestamp: dt.datetime,
        value: float,
    ) -> None:
        ts, uuid = ts_pair
        with raises(  # noqa: PT012
            TimeSeriesAddError,
            match="Error at upsert under DUPLICATE_POLICY == 'BLOCK'; got .*",
        ):
            for _ in range(2):
                _ = time_series_add(ts, f"{uuid}_{key}", timestamp, value)

    @given(
        ts_pair=redis_time_series(),
        key=text_ascii(),
        timestamp=zoned_datetimes(max_value=EPOCH_NAIVE).map(drop_microseconds),
        value=int32s() | floats(allow_nan=False, allow_infinity=False),
    )
    def test_invalid_timestamp(
        self,
        *,
        ts_pair: tuple[TimeSeries, UUID],
        key: str,
        timestamp: dt.datetime,
        value: float,
    ) -> None:
        _ = assume(timestamp < EPOCH_UTC)
        ts, uuid = ts_pair
        with raises(
            TimeSeriesAddError, match="Timestamp must be at least the Epoch; got .*"
        ):
            _ = time_series_add(
                ts, f"{uuid}_{key}", timestamp, value, duplicate_policy="last"
            )

    @given(
        ts_pair=redis_time_series(),
        key=text_ascii(),
        timestamp=datetimes(
            min_value=EPOCH_NAIVE, timezones=sampled_from([HONG_KONG, UTC])
        ).map(drop_microseconds),
    )
    @mark.parametrize("value", [param(inf), param(-inf), param(nan)])
    def test_invalid_value(
        self,
        *,
        ts_pair: tuple[TimeSeries, UUID],
        key: str,
        timestamp: dt.datetime,
        value: float,
    ) -> None:
        ts, uuid = ts_pair
        with raises(TimeSeriesAddError, match="Invalid value; got .*"):
            _ = time_series_add(
                ts,
                f"{uuid}_{key}",
                _clean_datetime(timestamp),
                value,
                duplicate_policy="last",
            )


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesAddAndReadDataFrame:
    schema: ClassVar[SchemaDict] = {
        "key": Utf8,
        "timestamp": DatetimeUTC,
        "value": Float64,
    }

    @given(
        data=data(),
        ts_pair=redis_time_series(),
        keys=lists_fixed_length(text_ascii(), 6, unique=True),
        datetime1=datetimes(min_value=EPOCH_NAIVE).map(drop_microseconds),
        datetime2=datetimes(min_value=EPOCH_NAIVE).map(drop_microseconds),
        time_zone=sampled_from([HONG_KONG, UTC]),
    )
    @mark.parametrize(
        ("strategy1", "dtype1"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    @mark.parametrize(
        ("strategy2", "dtype2"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    @settings(suppress_health_check={HealthCheck.filter_too_much})
    def test_main(
        self,
        *,
        data: DataObject,
        ts_pair: tuple[TimeSeries, UUID],
        keys: list[str],
        datetime1: dt.datetime,
        datetime2: dt.datetime,
        time_zone: ZoneInfo,
        strategy1: SearchStrategy[Number],
        dtype1: DataType,
        strategy2: SearchStrategy[Number],
        dtype2: DataType,
    ) -> None:
        key, id1, id2, timestamp, key_value1, key_value2 = keys
        ts, uuid = ts_pair
        full_id1, full_id2 = (f"{uuid}_{id_}" for id_ in [id1, id2])
        timestamp1, timestamp2 = (
            _clean_datetime(d, time_zone=time_zone) for d in [datetime1, datetime2]
        )
        value11, value21 = data.draw(tuples(strategy1, strategy1))
        value12, value22 = data.draw(tuples(strategy2, strategy2))
        schema = {
            key: Utf8,
            timestamp: zoned_datetime(time_zone=time_zone),
            key_value1: dtype1,
            key_value2: dtype2,
        }
        df = DataFrame(
            [
                (full_id1, timestamp1, value11, value12),
                (full_id2, timestamp2, value21, value22),
            ],
            schema=schema,
            orient="row",
        )
        time_series_add_dataframe(
            ts, df, key=key, timestamp=timestamp, duplicate_policy="last"
        )
        result = time_series_read_dataframe(
            ts,
            [full_id1, full_id2],
            [key_value1, key_value2],
            output_key=key,
            output_timestamp=timestamp,
            output_time_zone=time_zone,
        )
        check_polars_dataframe(result, height=2, schema_list=schema)
        assert_frame_equal(result, df)

    @given(ts_pair=redis_time_series())
    def test_error_add_key_missing(self, *, ts_pair: tuple[TimeSeries, UUID]) -> None:
        df = DataFrame()
        with raises(
            TimeSeriesAddDataFrameError,
            match="DataFrame must have a 'key' column; got .*",
        ):
            _ = time_series_add_dataframe(ts_pair[0], df)

    @given(ts_pair=redis_time_series())
    def test_error_add_timestamp_missing(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        df = DataFrame(schema={"key": Utf8})
        with raises(
            TimeSeriesAddDataFrameError,
            match="DataFrame must have a 'timestamp' column; got .*",
        ):
            _ = time_series_add_dataframe(ts_pair[0], df)

    @given(ts_pair=redis_time_series())
    def test_error_add_key_is_not_utf8(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        df = DataFrame(schema={"key": Boolean, "timestamp": DatetimeUTC})
        with raises(
            TimeSeriesAddDataFrameError,
            match="The 'key' column must be Utf8; got Boolean",
        ):
            _ = time_series_add_dataframe(ts_pair[0], df)

    @given(ts_pair=redis_time_series())
    def test_error_madd_timestamp_is_not_a_zoned_datetime(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": Boolean})
        with raises(
            TimeSeriesAddDataFrameError,
            match="The 'timestamp' column must be a zoned Datetime; got Boolean",
        ):
            _ = time_series_add_dataframe(ts_pair[0], df)

    @given(ts_pair=redis_time_series())
    def test_error_read_no_keys_requested(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        with raises(
            TimeSeriesReadDataFrameError, match="At least 1 key must be requested"
        ):
            _ = time_series_read_dataframe(ts_pair[0], [], [])

    @given(ts_pair=redis_time_series(), key=text_ascii())
    def test_error_read_no_columns_requested(
        self, *, ts_pair: tuple[TimeSeries, UUID], key: str
    ) -> None:
        ts, uuid = ts_pair
        with raises(
            TimeSeriesReadDataFrameError, match="At least 1 column must be requested"
        ):
            _ = time_series_read_dataframe(ts, f"{uuid}_{key}", [])


@SKIPIF_CI_AND_NOT_LINUX
class TestTimeSeriesMAddAndRange:
    int_schema: ClassVar[SchemaDict] = {
        "key": Utf8,
        "timestamp": DatetimeUTC,
        "value": Int64,
    }
    float_schema: ClassVar[SchemaDict] = {
        "key": Utf8,
        "timestamp": DatetimeUTC,
        "value": Float64,
    }

    @FLAKY
    @given(
        data=data(),
        ts_pair=redis_time_series(),
        keys=lists_fixed_length(text_ascii(), 5, unique=True),
        datetime1=datetimes(min_value=EPOCH_NAIVE).map(drop_microseconds),
        datetime2=datetimes(min_value=EPOCH_NAIVE).map(drop_microseconds),
        time_zone=sampled_from([HONG_KONG, UTC]),
    )
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    @mark.parametrize(
        ("strategy", "dtype"),
        [
            param(int32s(), Int64),
            param(floats(allow_nan=False, allow_infinity=False), Float64),
        ],
    )
    @settings(suppress_health_check={HealthCheck.filter_too_much})
    def test_main(
        self,
        *,
        data: DataObject,
        ts_pair: tuple[TimeSeries, UUID],
        case: Literal["values", "DataFrame"],
        keys: list[str],
        datetime1: dt.datetime,
        datetime2: dt.datetime,
        time_zone: ZoneInfo,
        strategy: SearchStrategy[Number],
        dtype: PolarsDataType,
    ) -> None:
        key, key1, key2, timestamp, value = keys
        ts, uuid = ts_pair
        full_keys = [f"{uuid}_{case}_{key}" for key in [key1, key2]]
        timestamps = [
            _clean_datetime(d, time_zone=time_zone) for d in [datetime1, datetime2]
        ]
        values = data.draw(tuples(strategy, strategy))
        triples = list(zip(full_keys, timestamps, values, strict=True))
        schema = {
            key: Utf8,
            timestamp: zoned_datetime(time_zone=time_zone),
            value: dtype,
        }
        match case:
            case "values":
                values_or_df = triples
            case "DataFrame":
                values_or_df = DataFrame(triples, schema=schema, orient="row")
        res_madd = time_series_madd(
            ts,
            values_or_df,
            key=key,
            timestamp=timestamp,
            value=value,
            duplicate_policy="last",
        )
        assert isinstance(res_madd, list)
        for i in res_madd:
            assert isinstance(i, int)
        res_range = time_series_range(
            ts,
            full_keys,
            output_key=key,
            output_timestamp=timestamp,
            output_time_zone=time_zone,
            output_value=value,
        )
        check_polars_dataframe(res_range, height=2, schema_list=schema)
        assert res_range.rows() == triples

    @given(ts_pair=redis_time_series())
    def test_error_madd_key_missing(self, *, ts_pair: tuple[TimeSeries, UUID]) -> None:
        ts, _ = ts_pair
        df = DataFrame()
        with raises(
            TimeSeriesMAddError, match="DataFrame must have a 'key' column; got .*"
        ):
            _ = time_series_madd(ts, df)

    @given(ts_pair=redis_time_series())
    def test_error_madd_timestamp_missing(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        ts, _ = ts_pair
        df = DataFrame(schema={"key": Utf8})
        with raises(
            TimeSeriesMAddError,
            match="DataFrame must have a 'timestamp' column; got .*",
        ):
            _ = time_series_madd(ts, df)

    @given(ts_pair=redis_time_series())
    def test_error_madd_value_missing(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        ts, _ = ts_pair
        df = DataFrame(schema={"key": Utf8, "timestamp": DatetimeUTC})
        with raises(
            TimeSeriesMAddError, match="DataFrame must have a 'value' column; got .*"
        ):
            _ = time_series_madd(ts, df)

    @given(ts_pair=redis_time_series())
    def test_error_madd_key_is_not_utf8(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        df = DataFrame(
            schema={"key": Boolean, "timestamp": DatetimeUTC, "value": Float64}
        )
        with raises(
            TimeSeriesMAddError, match="The 'key' column must be Utf8; got Boolean"
        ):
            _ = time_series_madd(ts_pair[0], df)

    @given(ts_pair=redis_time_series())
    def test_error_madd_timestamp_is_not_a_zoned_datetime(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": Boolean, "value": Float64})
        with raises(
            TimeSeriesMAddError,
            match="The 'timestamp' column must be a zoned Datetime; got Boolean",
        ):
            _ = time_series_madd(ts_pair[0], df)

    @given(ts_pair=redis_time_series())
    def test_error_madd_value_is_not_numeric(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        df = DataFrame(schema={"key": Utf8, "timestamp": DatetimeUTC, "value": Boolean})
        with raises(
            TimeSeriesMAddError, match="The 'value' column must be numeric; got Boolean"
        ):
            _ = time_series_madd(ts_pair[0], df)

    @FLAKY
    @given(
        ts_pair=redis_time_series(),
        key=text_ascii(),
        timestamp=zoned_datetimes(min_value=EPOCH_NAIVE).map(drop_microseconds),
        value=int32s(),
    )
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    def test_error_madd_invalid_key(
        self,
        *,
        ts_pair: tuple[TimeSeries, UUID],
        case: Literal["values", "DataFrame"],
        key: str,
        timestamp: dt.datetime,
        value: float,
    ) -> None:
        ts, uuid = ts_pair
        data = [(f"{uuid}_{case}_{key}", timestamp, value)]
        match case:
            case "values":
                values_or_df = data
            case "DataFrame":
                values_or_df = DataFrame(data, schema=self.int_schema, orient="row")
        with raises(TimeSeriesMAddError, match="The key '.*' must exist"):
            _ = time_series_madd(ts, values_or_df, assume_time_series_exist=True)

    @given(
        ts_pair=redis_time_series(),
        key=text_ascii(),
        timestamp=zoned_datetimes(max_value=EPOCH_NAIVE).map(drop_microseconds),
        value=int32s(),
    )
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    def test_error_madd_invalid_timestamp(
        self,
        *,
        ts_pair: tuple[TimeSeries, UUID],
        case: Literal["values", "DataFrame"],
        key: str,
        timestamp: dt.datetime,
        value: float,
    ) -> None:
        _ = assume(timestamp < EPOCH_UTC)
        ts, uuid = ts_pair
        data = [(f"{uuid}_{case}_{key}", timestamp, value)]
        match case:
            case "values":
                values_or_df = data
            case "DataFrame":
                values_or_df = DataFrame(data, schema=self.int_schema, orient="row")
        with raises(
            TimeSeriesMAddError, match="Timestamps must be at least the Epoch; got .*"
        ):
            _ = time_series_madd(ts, values_or_df)

    @given(
        ts_pair=redis_time_series(),
        key=text_ascii(),
        timestamp=datetimes(
            min_value=EPOCH_NAIVE, timezones=sampled_from([HONG_KONG, UTC])
        ).map(drop_microseconds),
    )
    @mark.parametrize("case", [param("values"), param("DataFrame")])
    @mark.parametrize("value", [param(inf), param(-inf), param(nan)])
    def test_error_madd_invalid_value(
        self,
        *,
        ts_pair: tuple[TimeSeries, UUID],
        case: Literal["values", "DataFrame"],
        key: str,
        timestamp: dt.datetime,
        value: float,
    ) -> None:
        ts, uuid = ts_pair
        data = [(f"{uuid}_{case}_{key}", _clean_datetime(timestamp), value)]
        match case:
            case "values":
                values_or_df = data
            case "DataFrame":
                values_or_df = DataFrame(data, schema=self.float_schema, orient="row")
        with raises(TimeSeriesMAddError, match="The value .* is invalid"):
            _ = time_series_madd(ts, values_or_df)

    @given(ts_pair=redis_time_series())
    def test_error_range_no_keys_requested(
        self, *, ts_pair: tuple[TimeSeries, UUID]
    ) -> None:
        ts, _ = ts_pair
        with raises(
            TimeSeriesRangeError, match="At least 1 key must be requested; got .*"
        ):
            _ = time_series_range(ts, [])

    @given(ts_pair=redis_time_series(), key=text_ascii())
    def test_error_range_invalid_key(
        self, *, ts_pair: tuple[TimeSeries, UUID], key: str
    ) -> None:
        ts, uuid = ts_pair
        with raises(TimeSeriesRangeError, match="The key '.*' must exist"):
            _ = time_series_range(ts, f"{uuid}_{key}")

    @given(
        ts_pair=redis_time_series(),
        key=text_ascii(),
        timestamp=zoned_datetimes(min_value=EPOCH_NAIVE).map(drop_microseconds),
        value=int32s(),
    )
    def test_error_range_key_with_int64_and_float64(
        self,
        *,
        ts_pair: tuple[TimeSeries, UUID],
        key: str,
        timestamp: dt.datetime,
        value: int,
    ) -> None:
        ts, uuid = ts_pair
        _ = time_series_madd(
            ts, [(f"{uuid}_{key}", timestamp, value)], duplicate_policy="last"
        )
        _ = time_series_madd(
            ts, [(f"{uuid}_{key}", timestamp, float(value))], duplicate_policy="last"
        )
        with raises(
            TimeSeriesRangeError,
            match="The key '.*' contains both Int64 and Float64 data",
        ):
            _ = time_series_range(ts, f"{uuid}_{key}")


class TestYieldClient:
    def test_sync(self) -> None:
        with yield_client() as client:
            assert isinstance(client, redis.Redis)

    async def test_async(self) -> None:
        async with yield_client_async() as client:
            assert isinstance(client, redis.asyncio.Redis)


class TestYieldTimeSeries:
    def test_sync(self) -> None:
        with yield_time_series() as ts:
            assert isinstance(ts, TimeSeries)

    async def test_async(self) -> None:
        async with yield_time_series_async() as ts:
            assert isinstance(ts, TimeSeries)



from __future__ import annotations

import datetime as dt
from typing import TYPE_CHECKING

from pytest import mark, param

from utilities.loguru import InterceptHandler, logged_sleep_async, logged_sleep_sync

if TYPE_CHECKING:
    from utilities.types import Duration


class TestInterceptHandler:
    def test_main(self) -> None:
        _ = InterceptHandler()


class TestLoggedSleep:
    @mark.parametrize("duration", [param(0.01), param(dt.timedelta(seconds=0.1))])
    def test_sync(self, *, duration: Duration) -> None:
        logged_sleep_sync(duration)

    @mark.parametrize("duration", [param(0.01), param(dt.timedelta(seconds=0.1))])
    async def test_async(self, *, duration: Duration) -> None:
        await logged_sleep_async(duration)



from __future__ import annotations

import datetime as dt
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum, StrEnum, unique
from fractions import Fraction
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypedDict,
    TypeVar,
    assert_never,
    cast,
    overload,
)

from orjson import (
    OPT_NON_STR_KEYS,
    OPT_PASSTHROUGH_DATETIME,
    OPT_SORT_KEYS,
    dumps,
    loads,
)
from typing_extensions import override

from utilities.dataclasses import Dataclass
from utilities.functions import get_class_name

if TYPE_CHECKING:
    from collections.abc import Callable

    from sqlalchemy.engine import Engine

_T = TypeVar("_T")
_TDataclass = TypeVar("_TDataclass", bound=Dataclass)
_SCHEMA_KEY = "_k"
_SCHEMA_VALUE = "_v"


@unique
class _Key(StrEnum):
    bytes = "byt"
    complex = "cmp"
    date = "dat"
    decimal = "dec"
    fraction = "frc"
    frozenset = "frz"
    ipv4_address = "ip4"
    ipv6_address = "ip6"
    local_datetime = "ldt"
    path = "pth"
    set = "set"
    slice = "slc"
    sqlalchemy_engine = "sql.eng"
    time = "tim"
    timedelta = "td"
    zoned_datetime = "zdt"


def serialize(obj: Any, /) -> bytes:
    """Serialize an object."""
    return dumps(
        obj,
        default=_serialize_default,
        option=OPT_NON_STR_KEYS | OPT_PASSTHROUGH_DATETIME | OPT_SORT_KEYS,
    )


class _SchemaDict(Generic[_T], TypedDict):
    _k: _Key
    _v: _T


def _serialize_default(obj: Any, /) -> _SchemaDict:
    schema = _get_schema(obj)
    return {_SCHEMA_KEY: schema.key, _SCHEMA_VALUE: schema.serializer(obj)}


@dataclass(frozen=True, kw_only=True)
class _Schema(Generic[_T]):
    key: _Key
    serializer: Callable[[_T], Any]


def _get_schema(obj: _T, /) -> _Schema[_T]:
    # standard library
    if isinstance(obj, bytes):
        return cast(_Schema[_T], _get_schema_bytes())
    if isinstance(obj, complex):
        return cast(_Schema[_T], _get_schema_complex())
    if isinstance(obj, dt.date) and not isinstance(obj, dt.datetime):
        return cast(_Schema[_T], _get_schema_date())
    if isinstance(obj, dt.datetime) and (obj.tzinfo is None):
        return cast(_Schema[_T], _get_schema_local_datetime())
    if isinstance(obj, dt.datetime) and (obj.tzinfo is not None):
        return cast(_Schema[_T], _get_schema_zoned_datetime())
    if isinstance(obj, Decimal):
        return cast(_Schema[_T], _get_schema_decimal())
    if isinstance(obj, Fraction):
        return cast(_Schema[_T], _get_schema_fraction())
    if isinstance(obj, IPv4Address):
        return cast(_Schema[_T], _get_schema_ipv4adress())
    if isinstance(obj, IPv6Address):
        return cast(_Schema[_T], _get_schema_ipv6adress())
    if isinstance(obj, Path):
        return cast(_Schema[_T], _get_schema_path())
    if isinstance(obj, slice):
        return cast(_Schema[_T], _get_schema_slice())
    if isinstance(obj, dt.time):
        return cast(_Schema[_T], _get_schema_time())
    if isinstance(obj, dt.timedelta):
        return cast(_Schema[_T], _get_schema_timedelta())
    # collections
    if isinstance(obj, frozenset):
        return cast(_Schema[_T], _get_schema_frozenset())
    if isinstance(obj, set):
        return cast(_Schema[_T], _get_schema_set())
    # third party
    if (schema := _get_schema_engine(obj)) is not None:
        return cast(_Schema[_T], schema)
    raise _GetSchemaError(obj=obj)


def _get_schema_bytes() -> _Schema[bytes]:
    return _Schema(key=_Key.bytes, serializer=lambda b: b.decode())


def _get_schema_complex() -> _Schema[complex]:
    return _Schema(key=_Key.complex, serializer=lambda c: (c.real, c.imag))


def _get_schema_date() -> _Schema[dt.date]:
    from utilities.whenever import serialize_date

    return _Schema(key=_Key.date, serializer=serialize_date)


def _get_schema_decimal() -> _Schema[Decimal]:
    return _Schema(key=_Key.decimal, serializer=str)


def _get_schema_engine(obj: Any, /) -> _Schema[Engine] | None:
    try:
        from sqlalchemy import Engine
    except ModuleNotFoundError:  # pragma: no cover
        pass
    else:
        if isinstance(obj, Engine):
            return _Schema(
                key=_Key.sqlalchemy_engine,
                serializer=lambda e: e.url.render_as_string(hide_password=False),
            )
    return None


def _get_schema_fraction() -> _Schema[Fraction]:
    return _Schema(key=_Key.fraction, serializer=lambda f: (f.numerator, f.denominator))


def _get_schema_frozenset() -> _Schema[frozenset[Any]]:
    def serializer(obj: frozenset[_T], /) -> list[_T]:
        try:
            return sorted(cast(Any, obj))
        except TypeError:
            return list(obj)

    return _Schema(key=_Key.frozenset, serializer=serializer)


def _get_schema_ipv4adress() -> _Schema[IPv4Address]:
    return _Schema(key=_Key.ipv4_address, serializer=str)


def _get_schema_ipv6adress() -> _Schema[IPv6Address]:
    return _Schema(key=_Key.ipv6_address, serializer=str)


def _get_schema_local_datetime() -> _Schema[dt.datetime]:
    from utilities.whenever import serialize_local_datetime

    return _Schema(key=_Key.local_datetime, serializer=serialize_local_datetime)


def _get_schema_path() -> _Schema[Path]:
    return _Schema(key=_Key.path, serializer=str)


def _get_schema_set() -> _Schema[set[Any]]:
    def serializer(obj: set[_T], /) -> list[_T]:
        try:
            return sorted(cast(Any, obj))
        except TypeError:
            return list(obj)

    return _Schema(key=_Key.set, serializer=serializer)


def _get_schema_slice() -> _Schema[slice]:
    return _Schema(key=_Key.slice, serializer=lambda s: (s.start, s.stop, s.step))


def _get_schema_time() -> _Schema[dt.time]:
    from utilities.whenever import serialize_time

    return _Schema(key=_Key.time, serializer=serialize_time)


def _get_schema_timedelta() -> _Schema[dt.timedelta]:
    from utilities.whenever import serialize_timedelta

    return _Schema(key=_Key.timedelta, serializer=serialize_timedelta)


def _get_schema_zoned_datetime() -> _Schema[dt.datetime]:
    from utilities.whenever import serialize_zoned_datetime

    return _Schema(key=_Key.zoned_datetime, serializer=serialize_zoned_datetime)


@dataclass(kw_only=True)
class _GetSchemaError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return f"Unsupported type: {get_class_name(self.obj)!r}"  # pragma: no cover


@overload
def deserialize(obj: bytes, /, *, cls: type[_TDataclass]) -> _TDataclass: ...
@overload
def deserialize(obj: bytes, /, *, cls: None = ...) -> Any: ...
def deserialize(obj: bytes, /, *, cls: type[_TDataclass] | None = None) -> Any:
    """Deserialize an object."""
    data = _object_hook(loads(obj))
    if cls is None:
        return data
    from dacite import Config, from_dict

    return from_dict(cls, data, config=Config(cast=[Enum]))


def _object_hook(obj: Any, /) -> Any:
    """Object hook for deserialization."""
    if not isinstance(obj, dict):
        return obj
    if set(obj) != {_SCHEMA_KEY, _SCHEMA_VALUE}:
        return {k: _object_hook(v) for k, v in obj.items()}
    schema: _SchemaDict[Any] = cast(Any, obj)
    value = schema[_SCHEMA_VALUE]
    match schema[_SCHEMA_KEY]:
        # standard library
        case _Key.bytes:
            return _object_hook_bytes(value)
        case _Key.complex:
            return _object_hook_complex(value)
        case _Key.date:
            return _object_hook_date(value)
        case _Key.decimal:
            return _object_hook_decimal(value)
        case _Key.fraction:
            return _object_hook_fraction(value)
        case _Key.ipv4_address:
            return _object_hook_ipv4_address(value)
        case _Key.ipv6_address:
            return _object_hook_ipv6_address(value)
        case _Key.local_datetime:
            return _object_hook_local_datetime(value)
        case _Key.path:
            return _object_hook_path(value)
        case _Key.slice:
            return _object_hook_slice(value)
        case _Key.time:
            return _object_hook_time(value)
        case _Key.timedelta:
            return _object_hook_timedelta(value)
        case _Key.zoned_datetime:
            return _object_hook_zoned_datetime(value)
        # collections
        case _Key.frozenset:
            return _object_hook_frozenset(value)
        case _Key.set:
            return _object_hook_set(value)
        # third party
        case _Key.sqlalchemy_engine:
            return _object_hook_sqlalchemy_engine(value)
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)


def _object_hook_bytes(value: str, /) -> bytes:
    return value.encode()


def _object_hook_complex(value: tuple[int, int], /) -> complex:
    real, imag = value
    return complex(real, imag)


def _object_hook_date(value: str, /) -> dt.date:
    from utilities.whenever import parse_date

    return parse_date(value)


def _object_hook_decimal(value: str, /) -> Decimal:
    return Decimal(value)


def _object_hook_fraction(value: tuple[int, int], /) -> Fraction:
    numerator, denominator = value
    return Fraction(numerator=numerator, denominator=denominator)


def _object_hook_frozenset(value: list[_T], /) -> frozenset[_T]:
    return frozenset(value)


def _object_hook_ipv4_address(value: str, /) -> IPv4Address:
    return IPv4Address(value)


def _object_hook_ipv6_address(value: str, /) -> IPv6Address:
    return IPv6Address(value)


def _object_hook_local_datetime(value: str, /) -> dt.date:
    from utilities.whenever import parse_local_datetime

    return parse_local_datetime(value)


def _object_hook_path(value: str, /) -> Path:
    return Path(value)


def _object_hook_set(value: list[_T], /) -> set[_T]:
    return set(value)


def _object_hook_slice(value: tuple[int | None, int | None, int | None], /) -> slice:
    start, stop, step = value
    return slice(start, stop, step)


def _object_hook_sqlalchemy_engine(value: str, /) -> Any:
    from sqlalchemy import create_engine

    return create_engine(value)


def _object_hook_time(value: str, /) -> dt.time:
    from utilities.whenever import parse_time

    return parse_time(value)


def _object_hook_timedelta(value: str, /) -> dt.timedelta:
    from utilities.whenever import parse_timedelta

    return parse_timedelta(value)


def _object_hook_zoned_datetime(value: str, /) -> dt.date:
    from utilities.whenever import parse_zoned_datetime

    return parse_zoned_datetime(value)


__all__ = ["deserialize", "serialize"]



from __future__ import annotations

from logging import getLogger

from pytest import mark, param, raises

from utilities.logging import (
    GetLoggingLevelError,
    LogLevel,
    basic_config,
    get_logging_level,
)


class TestBasicConfig:
    def test_main(self) -> None:
        basic_config()
        logger = getLogger(__name__)
        logger.info("message")


class TestGetLoggingLevel:
    @mark.parametrize(
        ("level", "expected"),
        [
            param(LogLevel.DEBUG, 10),
            param(LogLevel.INFO, 20),
            param(LogLevel.WARNING, 30),
            param(LogLevel.ERROR, 40),
            param(LogLevel.CRITICAL, 50),
        ],
    )
    def test_main(self, *, level: str, expected: int) -> None:
        assert get_logging_level(level) == expected

    def test_error(self) -> None:
        with raises(
            GetLoggingLevelError, match="Logging level 'invalid' must be valid"
        ):
            _ = get_logging_level("invalid")


class TestLogLevel:
    def test_main(self) -> None:
        assert len(LogLevel) == 5
